import logging
from random import randint
from typing import List

import numpy as np
import torch
from sklearn.model_selection import train_test_split
from torch.utils.data import Dataset
from tqdm import tqdm

from dbpunctuator.utils import NORMAL_TOKEN_TAG

PAD_TOKEN = "[PAD]"
DEFAULT_LABEL_WEIGHT = 0.5
logger = logging.getLogger(__name__)


def _read_data(file_path, min_sequence_length, max_sequence_length) -> List[List]:
    def read_line(text_line):
        return text_line.strip().split("\t")

    token_docs = []
    tag_docs = []
    line_index = 0

    token_doc = []
    tag_doc = []
    with open(file_path, "r") as data_file:
        pbar = tqdm(data_file.readlines())
        for line in pbar:
            if line_index == 0:
                token_doc = []
                tag_doc = []
                target_sequence_length = randint(
                    min_sequence_length, max_sequence_length
                )
            if line == "\n":
                token_docs.append(token_doc)
                tag_docs.append(tag_doc)
                line_index = 0
                pbar.update(len(token_doc))
                continue
            processed_line = read_line(line)
            try:
                token_doc.append(processed_line[0])
                tag_doc.append(processed_line[1])
            except IndexError:
                logger.warning(f"ignore the bad line: {line}")
                continue
            line_index += 1
            if line_index == target_sequence_length:
                token_docs.append(token_doc)
                tag_docs.append(tag_doc)
                line_index = 0
                pbar.update(target_sequence_length)
        token_doc += [PAD_TOKEN] * (target_sequence_length - line_index)
        tag_doc += [NORMAL_TOKEN_TAG] * (target_sequence_length - line_index)
        token_docs.append(token_doc)
        tag_docs.append(tag_doc)

        pbar.close()

    return token_docs, tag_docs


def generate_punctuator_tag_mappings(tag_docs):
    """
    Function for geneartion of label2id mapping if no mappings is provided

    Args:
        tag_docs(List[List[str]]): list of sequences of tags
    """
    all_tags = [tag for tags in tag_docs for tag in tags]
    unique_tags = np.unique(all_tags)
    label2id = {tag: id for id, tag in enumerate(unique_tags)}

    return label2id


def unison_shuffled_copies(a, b):
    assert len(a) == len(b)
    p = np.random.permutation(len(a))
    return a[p].tolist(), b[p].tolist()


def generate_evaluation_data(data_file_path, min_sequence_length, max_sequence_length):
    """
    Function for generation of evaluation assets including
    - evaluation corpus
    - evaluatio  tags

    Args:
        data_file_path(str): path of overall input data
        min_sequence_length(int): min sequence length of one sample
        max_sequence_length(int): max sequence length of one sample
    """
    logger.info("load validation data")
    texts, tags = _read_data(
        data_file_path,
        min_sequence_length,
        max_sequence_length,
    )
    return texts, tags


def generate_training_data(
    data_file_path, min_sequence_length, max_sequence_length, split_rate
):
    """
    Function for generation of training assets including
    - training corpus
    - training tags
    - validation corpus
    - validation tags

    Args:
        data_file_path(str): path of overall input data
        min_sequence_length(int): min sequence length of one sample
        max_sequence_length(int): max sequence length of one sample
        split_rate(float): train and validation split rate
    """
    logger.info("load training data")
    texts, tags = _read_data(
        data_file_path,
        min_sequence_length,
        max_sequence_length,
    )

    logger.info(f"data sample: {texts[0]}")
    (
        train_texts,
        val_texts,
        train_tags,
        val_tags,
    ) = train_test_split(texts, tags, test_size=split_rate, random_state=7)

    return train_texts, train_tags, val_texts, val_tags


class EncodingDataset(Dataset):
    def __init__(self, encodings, labels):
        self.encodings = encodings
        self.labels = labels

    def __getitem__(self, idx):
        item = {key: torch.tensor(val[idx]) for key, val in self.encodings.items()}
        item["labels"] = torch.tensor(self.labels[idx])
        return item

    def __len__(self):
        return len(self.labels)



import json
import logging
import re
from functools import wraps
from os import environ
from typing import Dict, Optional

import numpy as np
import torch
from plane.pattern import EMAIL, TELEPHONE
from pydantic import BaseModel
from transformers import DistilBertForTokenClassification, DistilBertTokenizerFast

from dbpunctuator.utils import (
    CURRENCY,
    CURRENCY_TOKEN,
    EMAIL_TOKEN,
    NUMBER,
    NUMBER_TOKEN,
    TELEPHONE_TOKEN,
    URL,
    URL_TOKEN,
    chinese_split,
    is_ascii,
)

logger = logging.getLogger(__name__)

num_regex = re.compile(f"{NUMBER.pattern}")
tel_regex = re.compile(f"{TELEPHONE.pattern}")
currency_regex = re.compile(f"{CURRENCY.pattern}")
email_regex = re.compile(f"{EMAIL.pattern}")
url_regex = re.compile(f"{URL.pattern}")


def verbose(attr_to_log):
    def wrapper_out(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            self = func(self, *args, **kwargs)
            if self.verbose:
                logger.debug(
                    f'After {func.__name__}, {attr_to_log} is generated as "{getattr(self, attr_to_log)}"'
                )
            return self

        return wrapper

    return wrapper_out


class InferenceArguments(BaseModel):
    """Arguments pertaining to which model/config/tokenizer we are going to fine-tune from.

    Args:
        model_name_or_path(str): name or path of pre-trained model
        tokenizer_name(str): name of pretrained tokenizer
        tag2punctuator(Dict[str, tuple]): tag to punctuator mapping.
            dbpunctuator.utils provides two mappings for English and Chinese
                NORMAL_TOKEN_TAG = "O"
                DEFAULT_ENGLISH_TAG_PUNCTUATOR_MAP = {
                    NORMAL_TOKEN_TAG: ("", False),
                    "COMMA": (",", False),
                    "PERIOD": (".", True),
                    "QUESTIONMARK": ("?", True),
                    "EXLAMATIONMARK": ("!", True),
                }

                DEFAULT_CHINESE_TAG_PUNCTUATOR_MAP = {
                    NORMAL_TOKEN_TAG: ("", False),
                    "C_COMMA": ("，", False),
                    "C_PERIOD": ("。", True),
                    "C_QUESTIONMARK": ("? ", True),
                    "C_EXLAMATIONMARK": ("! ", True),
                    "C_DUNHAO": ("、", False),
                }
            for own fine-tuned model with different tags, pass in your own mapping
        tag2id_storage_path(Optional[str]): tag2id storage path. Default one is from model config. Pass in this argument if your model doesn't have a tag2id inside config # noqa: E501
        gpu_device(int): specific gpu card index, default is the CUDA_VISIBLE_DEVICES from environ
    """

    model_name_or_path: str
    tokenizer_name: str
    tag2punctuator: Dict[str, tuple]
    tag2id_storage_path: Optional[str]
    gpu_device: int = environ.get("CUDA_VISIBLE_DEVICES", 0)


# whole pipeline running in the seperate process, provide a function for user to call, use socket for communication
class InferencePipeline:
    """Pipeline for inference"""

    def __init__(self, inference_arguments, verbose=False):
        if torch.cuda.is_available():
            self.device = torch.device(f"cuda:{inference_arguments.gpu_device}")
            logger.info(f"device type: {self.device.type}")
        else:
            self.device = torch.device("cpu")

        self.tokenizer = DistilBertTokenizerFast.from_pretrained(
            inference_arguments.tokenizer_name
        )
        self.classifier = DistilBertForTokenClassification.from_pretrained(
            inference_arguments.model_name_or_path
        ).to(self.device)
        if inference_arguments.tag2id_storage_path:
            with open(inference_arguments.tag2id_storage_path, "r") as fp:
                tag2id = json.load(fp)
                self.id2tag = {id: tag for tag, id in tag2id.items()}
        else:
            self.id2tag = self.classifier.config.id2label
        self.tag2punctuator = inference_arguments.tag2punctuator
        self.max_sequence_length = (
            self.classifier.config.max_position_embeddings // 2
        )  # for padding and subword

        self._reset_values()
        self.verbose = verbose

    @verbose("all_tokens")
    def pre_process(self, inputs):
        def _input_process(input_tokens):

            special_token_index = {}
            for index, token in enumerate(input_tokens):
                if email_regex.match(token):
                    input_tokens[index] = EMAIL_TOKEN
                    special_token_index[index] = token
                    continue
                if url_regex.match(token):
                    input_tokens[index] = URL_TOKEN
                    special_token_index[index] = token
                    continue
                if currency_regex.match(token):
                    input_tokens[index] = CURRENCY_TOKEN
                    special_token_index[index] = token
                    continue
                if tel_regex.match(token):
                    input_tokens[index] = TELEPHONE_TOKEN
                    special_token_index[index] = token
                    continue
                if num_regex.match(token):
                    input_tokens[index] = NUMBER_TOKEN
                    special_token_index[index] = token
                    continue
            return input_tokens, special_token_index

        index = 0
        last_is_split = False
        self.split_inputs_indexes = []
        for input in inputs:
            input_tokens = chinese_split(input).split()
            while len(input_tokens) > self.max_sequence_length:
                processed_input_tokens, special_token_index = list(
                    _input_process(input_tokens[: self.max_sequence_length])
                )
                self.special_token_indexes.append(special_token_index)
                self.all_tokens.append(processed_input_tokens)
                self.split_inputs_indexes.append(index)
                input_tokens = input_tokens[self.max_sequence_length :]
                index += 1
                last_is_split = True
            else:
                if last_is_split:
                    self.split_inputs_indexes.append(index)
                    last_is_split = False
                index += 1
                processed_input_tokens, special_token_index = list(
                    _input_process(input_tokens)
                )
                self.special_token_indexes.append(special_token_index)
                self.all_tokens.append(processed_input_tokens)
        logger.info(f"self split indexes: {self.split_inputs_indexes}")
        return self

    @verbose("tokenized_input_ids")
    def tokenize(self):
        tokenized_inputs = self.tokenizer(
            self.all_tokens,
            is_split_into_words=True,
            padding=True,
            return_offsets_mapping=True,
            return_tensors="pt",
        )
        self.marks = self._mark_ignored_tokens(tokenized_inputs["offset_mapping"])
        self.tokenized_input_ids = tokenized_inputs["input_ids"].to(self.device)
        self.attention_mask = tokenized_inputs["attention_mask"].to(self.device)
        return self

    @verbose("argmax_preds")
    def classify(self):
        try:
            logits = self.classifier(
                self.tokenized_input_ids, self.attention_mask
            ).logits
            if self.device.type == "cuda":
                self.argmax_preds = logits.argmax(dim=2).detach().cpu().numpy()
            else:
                self.argmax_preds = logits.argmax(dim=2).detach().numpy()
        except RuntimeError as e:
            logger.error(f"error doing punctuation: {str(e)}")
        return self

    @verbose("outputs")
    def post_process(self):
        reduce_ignored_marks = self.marks >= 0

        self.outputs_labels = []
        temp_ouputs = ""
        temp_outputs_labels = []
        for input_index, (
            pred,
            reduce_ignored,
            tokens,
            special_token_index,
        ) in enumerate(
            zip(
                self.argmax_preds,
                reduce_ignored_marks,
                self.all_tokens,
                self.special_token_indexes,
            )
        ):
            next_upper = True
            true_pred = pred[reduce_ignored]

            result_text = ""
            output_labels = []
            for id, (index, token) in zip(true_pred, enumerate(tokens)):
                tag = self.id2tag[id]
                output_labels.append(tag)
                if index in special_token_index:
                    token = special_token_index[index]
                if next_upper:
                    token = token.capitalize()
                punctuator, next_upper = self.tag2punctuator[tag]
                if is_ascii(token):
                    result_text += token + punctuator + " "
                else:
                    result_text += token + punctuator
            if input_index in self.split_inputs_indexes:
                temp_ouputs += result_text.strip()
                temp_outputs_labels.extend(output_labels)
            else:
                if temp_ouputs and temp_outputs_labels:
                    self.outputs.append(temp_ouputs.strip())
                    self.outputs_labels.append(temp_outputs_labels)
                    temp_ouputs = ""
                    temp_outputs_labels = []

                self.outputs.append(result_text.strip())
                self.outputs_labels.append(output_labels)

        if temp_ouputs and temp_outputs_labels:
            self.outputs.append(temp_ouputs.strip())
            self.outputs_labels.append(temp_outputs_labels)

        return self

    def punctuation(self, inputs):
        self._reset_values()
        self.pre_process(inputs).tokenize().classify().post_process()

        return self.outputs, self.outputs_labels

    def _mark_ignored_tokens(self, offset_mapping):
        samples = []
        for sample_offset in offset_mapping:
            # create an empty array of -100
            sample_marks = np.ones(len(sample_offset), dtype=int) * -100
            arr_offset = np.array(sample_offset)

            # set labels whose first offset position is 0 and the second is not 0, only special tokens second is also 0
            sample_marks[(arr_offset[:, 0] == 0) & (arr_offset[:, 1] != 0)] = 0
            samples.append(sample_marks.tolist())

        return np.array(samples)

    def _reset_values(self):
        self.special_token_indexes = []
        self.all_tokens = []
        self.outputs = []


class InferenceServer:
    """Inference server"""

    def __init__(
        self, inference_args, conn, termination, check_interval, verbose=False
    ) -> None:
        """Server for receiving tasks from client and do punctuation

        Args:
            server_address (str, optional): [server socket address]. Defaults to "/tmp/socket".
        """
        self.inference_pipeline = InferencePipeline(inference_args, verbose=verbose)
        self.conn = conn
        self.termination = termination
        self.check_interval = check_interval

    def punctuation(self):
        try:
            inputs = self.conn.recv()
            outputs_tuple = self.inference_pipeline.punctuation(inputs)
            self.conn.send(outputs_tuple)
        except OSError as err:
            logger.warning(f"error receiving inputs: {err}")

    def run(self):
        assert self.inference_pipeline, "no inference pipeline set up"
        logger.info("server is running")
        while True:
            try:
                if (
                    self.conn.poll(self.check_interval)
                    and not self.termination.is_set()
                ):
                    self.punctuation()
                if self.termination.is_set():
                    logger.info("termination is set")
                    break
            except OSError as err:
                logger.warning(f"sending output error: {err}")
                raise err
            except KeyboardInterrupt:
                logger.warning("punctuator shut down by keyboard interrupt")
                break
        self.terminate()

    def terminate(self):
        logger.info("terminate punctuation server")

        self.conn.close()



import logging
import time
from os import environ
from typing import Dict, List, Optional

import numpy as np
import torch
import torch.nn.functional as F
from pydantic import BaseModel
from sklearn.utils import class_weight
from torch.utils.data import DataLoader
from torch.utils.tensorboard import SummaryWriter
from tqdm import tqdm
from transformers import (
    AdamW,
    DistilBertConfig,
    DistilBertForTokenClassification,
    DistilBertTokenizerFast,
    get_constant_schedule_with_warmup,
)

logger = logging.getLogger(__name__)

DEFAULT_LABEL_WEIGHT = 1


class EncodingDataset:
    def __init__(self, encodings, labels):
        self.encodings = encodings
        self.labels = labels

    def __getitem__(self, idx):
        item = {key: torch.tensor(val[idx]) for key, val in self.encodings.items()}
        item["labels"] = torch.tensor(self.labels[idx])
        return item

    def __len__(self):
        return len(self.labels)


class NERTrainingArguments(BaseModel):
    """Arguments pertaining to which model/config/tokenizer we are going to fine-tune from.

    Args:
        # basic arguments
        training_corpus(List[List[str]]): list of sequences for training, longest sequence should be no longer than pretrained LM # noqa: E501
        validation_corpus(List[List[str]]): list of sequences for validation, longest sequence should be no longer than pretrained LM # noqa: E501
        training_tags(List[List[int]]): tags(int) for training
        validation_tags(List[List[int]]): tags(int) for validation
        model_name_or_path(str): name or path of pre-trained model
        tokenizer_name(str): name of pretrained tokenizer

        # training arguments
        epoch(int): number of epoch
        batch_size(int): batch size
        model_storage_dir(str): fine-tuned model storage path
        label2id(Dict): the tags label and id mapping
        early_stop_count(int): after how many epochs to early stop training if valid loss not become smaller. default 3 # noqa: E501
        gpu_device(int): specific gpu card index, default is the CUDA_VISIBLE_DEVICES from environ
        warm_up_steps(int): warm up steps.
        r_drop(bool): whether to train with r-drop
        r_alpha(int): alpha value for kl divengence in the loss, default is 0
        plot_steps(int): record training status to tensorboard among how many steps
        tensorboard_log_dir(Optional[str]): the tensorboard logs output directory, default is "runs"

        # model arguments
        addtional_model_config(Optional[Dict]): additional configuration for model
    """

    # basic args
    training_corpus: List[List[str]]
    validation_corpus: List[List[str]]
    training_tags: List[List[int]]
    validation_tags: List[List[int]]
    model_name_or_path: str
    tokenizer_name: str

    # training ars
    epoch: int
    batch_size: int
    model_storage_dir: str
    label2id: Dict
    early_stop_count: Optional[int] = 3
    gpu_device: int = environ.get("CUDA_VISIBLE_DEVICES", 0)
    warm_up_steps: int = 1000
    r_drop: bool = False
    r_alpha: int = 0
    plot_steps: int = 50
    tensorboard_log_dir: Optional[str] = "runs"

    # model args
    addtional_model_config: Optional[Dict]


class NERTrainingPipeline:
    def __init__(self, training_arguments):
        """Training pipeline for fine-tuning the distilbert token classifier for punctuation

        Args:
            training_arguments (TrainingArguments): arguments passed to training pipeline
        """
        self.arguments = training_arguments
        if torch.cuda.is_available():
            self.device = torch.device(f"cuda:{training_arguments.gpu_device}")
        else:
            self.device = torch.device("cpu")
        self.label2id = training_arguments.label2id
        self.id2label = {id: label for label, id in self.label2id.items()}
        self.tensorboard_writter = SummaryWriter(training_arguments.tensorboard_log_dir)

    def tokenize(self):
        logger.info("tokenize data")
        self.model_config = DistilBertConfig.from_pretrained(
            self.arguments.model_name_or_path,
            label2id=self.label2id,
            id2label=self.id2label,
            num_labels=len(self.id2label),
            **self.arguments.addtional_model_config,
        )
        tokenizer = DistilBertTokenizerFast.from_pretrained(
            self.arguments.tokenizer_name,
        )
        self.train_encodings = tokenizer(
            self.arguments.training_corpus,
            is_split_into_words=True,
            return_offsets_mapping=True,
            padding=True,
        )
        self.val_encodings = tokenizer(
            self.arguments.validation_corpus,
            is_split_into_words=True,
            return_offsets_mapping=True,
            padding=True,
        )

        all_ner_tag_ids = [
            tag_id
            for sen_tag_ids in self.arguments.training_tags
            + self.arguments.validation_tags
            for tag_id in sen_tag_ids
        ]
        unique_tag_ids = set(all_ner_tag_ids)

        logger.info(f"unique tag ids: {unique_tag_ids}, id2label: {self.id2label}")

        weights = [
            weight if weight > 0 else DEFAULT_LABEL_WEIGHT
            for weight in np.log(
                class_weight.compute_class_weight(
                    "balanced", classes=list(unique_tag_ids), y=all_ner_tag_ids
                )
            )
        ]
        logger.info(
            f"class weights: {[round(weight, 2) for weight in weights]}, id2label: {self.id2label}"
        )

        self.class_weights = torch.tensor(weights, dtype=torch.float)

        self.train_encoded_tags = self._encode_tags(
            self.arguments.training_tags, self.train_encodings
        )
        self.validation_encoded_tags = self._encode_tags(
            self.arguments.validation_tags, self.val_encodings
        )

        return self

    def generate_dataset(self):
        logger.info("generate dataset from tokenized data")
        self.train_encodings.pop("offset_mapping")
        self.val_encodings.pop("offset_mapping")
        self.training_dataset = EncodingDataset(
            self.train_encodings, self.train_encoded_tags
        )
        self.val_dataset = EncodingDataset(
            self.val_encodings, self.validation_encoded_tags
        )

        return self

    def fine_tune(self):
        logger.info("start fine tune")
        self.classifier = DistilBertForTokenClassification.from_pretrained(
            self.arguments.model_name_or_path,
            config=self.model_config,
        )
        self.classifier.to(self.device)

        train_loader = DataLoader(
            self.training_dataset, batch_size=self.arguments.batch_size, shuffle=True
        )
        val_loader = DataLoader(
            self.val_dataset, batch_size=self.arguments.batch_size, shuffle=True
        )
        optim = AdamW(self.classifier.parameters(), lr=1e-5)

        scheduler = get_constant_schedule_with_warmup(
            optim,
            num_warmup_steps=self.arguments.warm_up_steps,
        )

        best_valid_loss = 100
        no_improvement_count = 0
        self.total_steps = 0

        with tqdm(total=self.arguments.epoch) as pbar:
            for epoch in range(self.arguments.epoch):
                pbar.set_description(f"Processing epoch: {epoch + 1}")

                start_time = time.time()

                self.classifier.train()

                train_loss, train_acc = self._train(train_loader, optim, scheduler)
                val_loss, val_acc = self._train(val_loader, optim, scheduler, True)

                self.tensorboard_writter.add_scalar(
                    "Epoch Loss/train", train_loss, epoch + 1
                )
                self.tensorboard_writter.add_scalar(
                    "Epoch Loss/valid", val_loss, epoch + 1
                )

                self.tensorboard_writter.add_scalar(
                    "Epoch acc/train", train_acc, epoch + 1
                )
                self.tensorboard_writter.add_scalar(
                    "Epoch acc/valid", val_acc, epoch + 1
                )

                end_time = time.time()

                epoch_mins, epoch_secs = self._epoch_time(start_time, end_time)

                logger.info(
                    f"Epoch: {epoch + 1:02} | Epoch Time: {epoch_mins}m {epoch_secs}s"
                )
                logger.info(
                    f"\tTrain Loss: {train_loss:.3f} | Train Acc: {train_acc * 100:.2f}%"
                )
                logger.info(
                    f"\t Val. Loss: {val_loss:.3f} |  Val. Acc: {val_acc * 100:.2f}%"
                )
                pbar.update(1)

                if val_loss < best_valid_loss:
                    best_valid_loss = val_loss
                    self.best_state_dict = self.classifier.state_dict()
                    no_improvement_count = 0
                else:
                    no_improvement_count += 1
                    if (
                        self.arguments.early_stop_count > 0
                        and no_improvement_count >= self.arguments.early_stop_count
                    ):
                        logger.info(
                            f"No improvement for past {no_improvement_count} epochs, early stop training."
                        )
                        return self

            logger.info("fine-tune finished")

        self.tensorboard_writter.flush()

        return self

    def persist(self):
        """Persist this model into the passed directory."""

        logger.info("persist fine-tuned model")

        self.classifier.load_state_dict(self.best_state_dict)
        self.classifier.save_pretrained(self.arguments.model_storage_dir)

        logger.info(f"fine-tuned model stored to {self.arguments.model_storage_dir}")

    def _encode_tags(self, tags, encodings):
        logger.info("encoding tags")
        encoded_labels = []
        with tqdm(total=len(tags)) as pbar:
            for doc_labels, doc_offset in zip(tags, encodings.offset_mapping):
                try:
                    # create an empty array of -100
                    doc_enc_labels = np.ones(len(doc_offset), dtype=int) * -100
                    arr_offset = np.array(doc_offset)

                    # set labels whose first offset position is 0 and the second is not 0
                    doc_enc_labels[
                        (arr_offset[:, 0] == 0) & (arr_offset[:, 1] != 0)
                    ] = doc_labels
                    encoded_labels.append(doc_enc_labels.tolist())
                except ValueError as e:
                    logger.warning(f"error encoding: {str(e)}")
                pbar.update(1)

        return encoded_labels

    def _compute_kl_loss(self, p, q, pad_mask=None):
        p_loss = F.kl_div(
            F.log_softmax(p, dim=-1), F.softmax(q, dim=-1), reduction="none"
        )
        q_loss = F.kl_div(
            F.log_softmax(q, dim=-1),
            F.softmax(p, dim=-1),
            reduction="none",
        )

        # pad_mask is for seq-level tasks
        if pad_mask is not None:
            p_loss.masked_fill_(pad_mask, 0.0)
            q_loss.masked_fill_(pad_mask, 0.0)

        # # You can choose whether to use function "sum" and "mean" depending on your task
        p_loss = p_loss.sum()
        q_loss = q_loss.sum()

        loss = (p_loss + q_loss) / 2
        return loss

    def _train(self, iterator, optim, scheduler=None, is_val=False):
        epoch_loss = 0
        epoch_acc = 0
        if is_val:
            self.classifier.train(False)
        else:
            self.classifier.train()

        in_epoch_steps = 0

        with tqdm(total=len(iterator)) as pbar:
            for batch in iterator:
                in_epoch_steps += 1
                pbar.set_description(f"Processing batch: {in_epoch_steps}")

                optim.zero_grad()
                input_ids = batch["input_ids"].to(self.device)
                attention_mask = batch["attention_mask"].to(self.device)
                labels = batch["labels"].to(self.device)

                if self.arguments.r_drop:

                    outputs_1 = self.classifier(
                        input_ids, attention_mask=attention_mask, labels=labels
                    )
                    logits_1 = outputs_1.logits
                    if in_epoch_steps == 1:
                        logger.info(f"logits shape {logits_1.size()}")
                    logits_1 = logits_1.view(-1, logits_1.size(-1))
                    if in_epoch_steps == 1:
                        logger.info(f"viewed logits shape {logits_1.size()}")
                    loss_1 = F.cross_entropy(
                        logits_1,
                        labels.view(-1),
                        weight=self.class_weights.to(self.device),
                        reduction="mean",
                    )

                    outputs_2 = self.classifier(
                        input_ids, attention_mask=attention_mask, labels=labels
                    )
                    logits_2 = outputs_2.logits
                    logits_2 = logits_2.view(-1, logits_2.size(-1))
                    loss_2 = F.cross_entropy(
                        logits_2,
                        labels.view(-1),
                        weight=self.class_weights.to(self.device),
                        reduction="mean",
                    )

                    # cross entropy loss for classifier
                    ce_loss = 0.5 * (loss_1 + loss_2)
                    kl_loss = self._compute_kl_loss(logits_1, logits_2)

                    # carefully choose hyper-parameters
                    loss = ce_loss + self.arguments.r_alpha * kl_loss
                    logits = logits_1.add(logits_2) / 2  # average over two logits

                else:
                    outputs = self.classifier(
                        input_ids, attention_mask=attention_mask, labels=labels
                    )
                    logits = outputs.logits
                    logits = logits.view(-1, logits.size(-1))
                    loss = F.cross_entropy(
                        logits,
                        labels.view(-1),
                        # weight=self.class_weights.to(self.device),
                        reduction="mean",
                    )

                if not is_val:
                    loss.backward()
                    optim.step()
                    if scheduler:
                        scheduler.step()
                    if self.total_steps % self.arguments.plot_steps == 0:
                        self.tensorboard_writter.add_scalar(
                            "Step Loss/train", loss, self.total_steps
                        )

                self.total_steps += 1

                epoch_loss += loss.item()
                epoch_acc += self._accuracy(logits, attention_mask, labels)

                pbar.update(1)
                pbar.set_postfix(
                    {
                        "Last_loss": f"{loss:.2f}",
                        "Avg_cum_loss": f"{epoch_loss/in_epoch_steps:.2f}",
                    }
                )

        return epoch_loss / in_epoch_steps, epoch_acc / in_epoch_steps

    def _epoch_time(self, start_time, end_time):
        elapsed_time = end_time - start_time
        elapsed_mins = int(elapsed_time // 60)
        elapsed_secs = int(elapsed_time - (elapsed_mins * 60))
        return elapsed_mins, elapsed_secs

    def _accuracy(self, logits, attention_mask, labels):
        if self.device.type == "cuda":
            max_preds = logits.argmax(dim=-1).detach().cpu().numpy().flatten()
            flattened_labels = labels.detach().cpu().numpy().flatten()
            flattened_attention = attention_mask.detach().cpu().numpy().flatten()
        else:
            max_preds = logits.argmax(dim=-1).detach().numpy().flatten()
            flattened_labels = labels.detach().numpy().flatten()
            flattened_attention = attention_mask.detach().numpy().flatten()

        not_padding_labels = flattened_labels[flattened_attention == 1]
        not_padding_preds = max_preds[flattened_attention == 1]
        reduce_ignored = not_padding_labels >= 0
        true_labels = not_padding_labels[reduce_ignored]  # remove ignored -100
        true_preds = not_padding_preds[reduce_ignored]

        if true_preds.shape[0] == true_labels.shape[0]:
            return np.sum(true_preds == true_labels) / true_preds.shape[0]

        return 0

    def run(self):
        self.tokenize().generate_dataset().fine_tune().persist()



from .additional_data_process import (
    chinese_split,
    keep_only_latin_characters,
    remove_brackets_text,
)
from .constant import (
    ALL_PUNCS,
    CURRENCY,
    CURRENCY_TOKEN,
    DEFAULT_CHINESE_NER_MAPPING,
    DEFAULT_CHINESE_TAG_PUNCTUATOR_MAP,
    DEFAULT_ENGLISH_NER_MAPPING,
    DEFAULT_ENGLISH_TAG_PUNCTUATOR_MAP,
    EMAIL_TOKEN,
    NORMAL_TOKEN_TAG,
    NUMBER,
    NUMBER_TOKEN,
    TELEPHONE_TOKEN,
    URL,
    URL_TOKEN,
)
from .utils import is_ascii

__all__ = [
    "ALL_PUNCS",
    "DEFAULT_ENGLISH_NER_MAPPING",
    "DEFAULT_CHINESE_TAG_PUNCTUATOR_MAP",
    "DEFAULT_CHINESE_NER_MAPPING",
    "DEFAULT_ENGLISH_TAG_PUNCTUATOR_MAP",
    "NORMAL_TOKEN_TAG",
    "is_ascii",
    "URL",
    "CURRENCY",
    "EMAIL_TOKEN",
    "URL_TOKEN",
    "TELEPHONE_TOKEN",
    "CURRENCY_TOKEN",
    "NUMBER_TOKEN",
    "NUMBER",
    "remove_brackets_text",
    "keep_only_latin_characters",
    "chinese_split",
]



import logging
from os import environ
from typing import Dict, List, Optional

import numpy as np
import torch
from pydantic import BaseModel
from sklearn.metrics import classification_report
from torch._C import device  # noqa: F401
from torch.utils.data import DataLoader
from tqdm import tqdm
from transformers import DistilBertForTokenClassification, DistilBertTokenizerFast

from dbpunctuator.utils import NORMAL_TOKEN_TAG

from .punctuation_data_process import EncodingDataset

logger = logging.getLogger(__name__)


class EvaluationArguments(BaseModel):
    """Arguments pertaining to which model/config we are going to do the validation

    Args:
        evaluation_corpus(List[List[str]]): list of sequences for evaluation, longest sequence should be no longer than pretrained LM's max_position_embedding(512) # noqa: E501
        evaluation_tags(List[List[int]]): tags(int) for evaluation (the GT)
        model_name_or_path(str): name or path of fine-tuned model
        tokenizer_name(str): name of tokenizer
        batch_size(int): batch size
        label2id(Optional[Dict]): label2id. Default one is from model config. Pass in this argument if your model doesn't have a label2id inside config # noqa: E501
        gpu_device(int): specific gpu card index, default is the CUDA_VISIBLE_DEVICES from environ
    """

    evaluation_corpus: List[List[str]]
    evaluation_tags: List[List[int]]
    model_name_or_path: str
    tokenizer_name: str
    batch_size: int
    label2id: Optional[Dict]
    gpu_device: int = environ.get("CUDA_VISIBLE_DEVICES", 0)


class EvaluationPipeline:
    def __init__(self, validation_arguments) -> None:
        self.arguments = validation_arguments

        if torch.cuda.is_available():
            self.device = torch.device(f"cuda:{validation_arguments.gpu_device}")
        else:
            self.device = torch.device("cpu")

        self.tokenizer = DistilBertTokenizerFast.from_pretrained(
            validation_arguments.tokenizer_name
        )
        self.classifier = DistilBertForTokenClassification.from_pretrained(
            validation_arguments.model_name_or_path
        ).to(self.device)
        if validation_arguments.label2id:
            self.label2id = validation_arguments.label2id
        else:
            self.label2id = self.classifier.config.label2id

    def tokenize(self):
        logger.info("tokenize data")
        tokenizer = DistilBertTokenizerFast.from_pretrained(
            self.arguments.tokenizer_name
        )
        self.encodings = tokenizer(
            self.arguments.evaluation_corpus,
            is_split_into_words=True,
            return_offsets_mapping=True,
            padding=True,
        )
        self.evaluation_encoded_tags = self._encode_tags(
            self.arguments.evaluation_tags, self.encodings
        )
        self.dataset = EncodingDataset(self.encodings, self.evaluation_encoded_tags)

        return self

    def validate(self):
        logger.info("start validation")
        val_loader = DataLoader(
            self.dataset, batch_size=self.arguments.batch_size, shuffle=True
        )
        self.classifier.train(False)

        steps = 0
        total_preds = []
        total_labels = []
        with tqdm(total=len(val_loader)) as pbar:
            for batch in val_loader:
                steps += 1
                pbar.set_description(f"Processing batch: {steps}")

                input_ids = batch["input_ids"].to(self.device)
                attention_mask = batch["attention_mask"].to(self.device)
                labels = batch["labels"].to(self.device)
                outputs = self.classifier(
                    input_ids, attention_mask=attention_mask, labels=labels
                )
                loss = outputs.loss
                logits = outputs.logits

                true_preds, true_labels = self._post_process(
                    logits, labels, attention_mask
                )
                total_preds.extend(true_preds)
                total_labels.extend(true_labels)

                pbar.update(1)
                pbar.set_postfix(
                    {
                        "Last_batch_loss": f"{loss:.2f}",
                    }
                )

        tested_labels = []
        target_names = []
        for label, id in self.label2id.items():
            if label != NORMAL_TOKEN_TAG:
                tested_labels.append(id)
                target_names.append(label)
        report = classification_report(
            total_labels,
            total_preds,
            labels=tested_labels,
            digits=4,
            target_names=target_names,
            zero_division=1,
        )
        logger.info(f"validation report: \n {report}")

    def run(self):
        self.tokenize().validate()

    def _encode_tags(self, tags, encodings):
        logger.info("encoding tags")
        encoded_labels = []
        with tqdm(total=len(tags)) as pbar:
            for doc_labels, doc_offset in zip(tags, encodings.offset_mapping):
                try:
                    # create an empty array of -100
                    doc_enc_labels = np.ones(len(doc_offset), dtype=int) * -100
                    arr_offset = np.array(doc_offset)

                    # set labels whose first offset position is 0 and the second is not 0
                    doc_enc_labels[
                        (arr_offset[:, 0] == 0) & (arr_offset[:, 1] != 0)
                    ] = doc_labels
                    encoded_labels.append(doc_enc_labels.tolist())
                except ValueError as e:
                    logger.warning(f"error encoding: {str(e)}")
                pbar.update(1)

        return encoded_labels

    def _post_process(self, logits, labels, attention_mask):
        if self.device.type == "cuda":
            max_preds = logits.argmax(dim=2).detach().cpu().numpy().flatten()
            flattened_labels = labels.detach().cpu().numpy().flatten()
            flattened_attention = attention_mask.detach().cpu().numpy().flatten()
        else:
            max_preds = logits.argmax(dim=2).detach().numpy().flatten()
            flattened_labels = labels.detach().numpy().flatten()
            flattened_attention = attention_mask.detach().numpy().flatten()
        not_padding_labels = flattened_labels[flattened_attention == 1]
        not_padding_preds = max_preds[flattened_attention == 1]
        reduce_ignored = not_padding_labels >= 0
        true_labels = not_padding_labels[reduce_ignored]  # remove ignored -100
        true_preds = not_padding_preds[reduce_ignored]

        return true_preds, true_labels



from .evalute import EvaluationArguments, EvaluationPipeline
from .general_ner_train import NERTrainingArguments, NERTrainingPipeline
from .punctuation_data_process import generate_evaluation_data, generate_training_data

__all__ = [
    "EvaluationArguments",
    "EvaluationPipeline",
    "NERTrainingArguments",
    "NERTrainingPipeline",
    "generate_training_data",
    "generate_evaluation_data",
]



import logging
import os
from typing import List

import pandas as pd
from tqdm import tqdm

from dbpunctuator.utils import ALL_PUNCS, DEFAULT_ENGLISH_NER_MAPPING, NORMAL_TOKEN_TAG

from .data_cleanning import (
    cleaning_validator,
    dataframe_data_cleaning,
    text_lines_cleaning,
)

logger = logging.getLogger(__name__)


def cleanup_data_from_csv(
    source_data,
    target_col,
    output_file_path,
    ner_mapping=DEFAULT_ENGLISH_NER_MAPPING,
    additional_to_remove=[],
    special_cleaning_funcs=[],
):
    """clean up training data from csv file

    Args:
        source_data (string or Pandas.Dataframe): path of training data csv or own defined dataframe
        target_col (list or string): columns of training data in csv
        output_file_path (string): path of cleaned data
        ner_mapping (dict, optional): NER mapping of punctuation marks. Defaults to utils.constant.DEFAULT_ENGLISH_NER_MAPPING keys  # noqa: E501
        additional_to_remove (list, optional): additional special characters to remove, default []
        special_cleaning_funcs (List[funcs], optional): additional cleaning funcs to apply to csv data, default []
    """
    target_col = target_col if isinstance(target_col, list) else [target_col]
    if isinstance(source_data, pd.DataFrame):
        dataframe = source_data
    else:
        dataframe = pd.read_csv(source_data).dropna(subset=target_col)
    kept_punctuations = [ord(p) for p in set(ner_mapping.keys())]
    removed_punctuations = [p for p in ALL_PUNCS if p not in kept_punctuations] + [
        ord(p) for p in additional_to_remove
    ]
    logger.info("clean up original data")
    try:
        os.remove(output_file_path)
    except FileNotFoundError:
        pass
    for col in target_col:
        result_df = dataframe_data_cleaning(
            dataframe,
            col,
            kept_punctuations,
            removed_punctuations,
            *special_cleaning_funcs,
        )
        with open(output_file_path, "a+") as output_file:
            for row in result_df[col].tolist():
                try:
                    if row and cleaning_validator(
                        row, kept_punctuations, removed_punctuations
                    ):
                        output_file.write(row + " \n")
                except AssertionError as e:
                    logger.warning(str(e))


def clean_up_data_from_txt(
    source_data,
    output_file_path,
    ner_mapping=DEFAULT_ENGLISH_NER_MAPPING,
    additional_to_remove=[],
    special_cleaning_funcs=[],
):
    """clean up training data from text file

    Args:
        source_data (string or List): path of training data or own defined data
        output_file_path (string): path of cleaned data
        ner_mapping (dict, optional): NER mapping of punctuation marks. Defaults to utils.constant.DEFAULT_ENGLISH_NER_MAPPING keys  # noqa: E501
        additional_to_remove (list, optional): additional special characters to remove, default []
        special_cleaning_funcs (List[funcs], optional): additional cleaning funcs to apply to csv data, default []
    """
    kept_punctuations = [ord(p) for p in set(ner_mapping.keys())]
    removed_punctuations = [p for p in ALL_PUNCS if p not in kept_punctuations] + [
        ord(p) for p in additional_to_remove
    ]
    if isinstance(source_data, List):
        cleaned_up_lines = list(
            text_lines_cleaning(
                source_data,
                kept_punctuations,
                removed_punctuations,
                *special_cleaning_funcs,
            )
        )
    else:
        with open(source_data, "r") as file:
            cleaned_up_lines = list(
                text_lines_cleaning(
                    file.readlines(),
                    kept_punctuations,
                    removed_punctuations,
                    *special_cleaning_funcs,
                )
            )
    try:
        os.remove(output_file_path)
    except FileNotFoundError:
        pass
    with open(output_file_path, "a+") as output_file:
        for line in cleaned_up_lines:
            try:
                if line and cleaning_validator(
                    line, kept_punctuations, removed_punctuations
                ):
                    output_file.write(line + "\n")
            except AssertionError as e:
                logger.warning(str(e))


def process_line(line, ner_mapping):
    text_list = line.split()
    token_list = []
    tag_list = []
    if len(text_list) == 0:
        return token_list, tag_list
    # clean up puncs in the beginning of the text
    latest_word = text_list.pop(0)
    while latest_word in ner_mapping:
        if not text_list:
            break
        latest_word = text_list.pop(0)
    latest_token = NORMAL_TOKEN_TAG
    latest_is_punc = False
    for word in text_list:
        if word in ner_mapping:
            if not latest_is_punc:
                latest_token = ner_mapping[word]
                latest_is_punc = True
                token_list.append(latest_word)
                tag_list.append(latest_token)
            else:
                pass
        else:
            if not latest_is_punc:
                token_list.append(latest_word)
                tag_list.append(latest_token)
            latest_is_punc = False
            latest_word = word
            latest_token = NORMAL_TOKEN_TAG
    if not latest_is_punc:
        token_list.append(latest_word)
        tag_list.append(latest_token)
    return token_list, tag_list


def generate_corpus(
    cleaned_data_path,
    training_data_path,
    ner_mapping=DEFAULT_ENGLISH_NER_MAPPING,
):
    """generate "token tag" format training data based on cleaned text

    Args:
        cleaned_data_path (string): path of cleaned data
        training_data_path (string): path of generated training data
        ner_mapping (dict, optional): ner mapping for puncs and labels
    """
    logger.info("generate training data")
    with open(cleaned_data_path, "r") as data_file:
        lines = data_file.readlines()
    with open(training_data_path, "w+") as training_data_file:
        pbar = tqdm(lines)
        for line in pbar:
            tokens, tags = process_line(line, ner_mapping)
            for token, tag in zip(tokens, tags):
                training_data_file.write("%s\t%s\n" % (token, tag))
            training_data_file.write("\n")
        pbar.close()



import sys
import unicodedata

from plane import build_new_regex

NORMAL_TOKEN_TAG = "O"
EMAIL_TOKEN = "<EMAIL>"
URL_TOKEN = "<URL>"
TELEPHONE_TOKEN = "<TEL>"
CURRENCY_TOKEN = "<CURRENCY>"
NUMBER_TOKEN = "<NUM>"
URL = build_new_regex(
    "url_checking",
    r"https?:\/\/[!-~]+|[!-~]+\.[-_a-z/]+",
)


DEFAULT_ENGLISH_TAG_PUNCTUATOR_MAP = {
    NORMAL_TOKEN_TAG: ("", False),
    "COMMA": (",", False),
    "PERIOD": (".", True),
    "QUESTIONMARK": ("?", True),
    "EXLAMATIONMARK": ("!", True),
}

DEFAULT_CHINESE_TAG_PUNCTUATOR_MAP = {
    NORMAL_TOKEN_TAG: ("", False),
    "C_COMMA": ("，", False),
    "C_PERIOD": ("。", True),
    "C_QUESTIONMARK": ("? ", True),
    "C_EXLAMATIONMARK": ("! ", True),
    "C_DUNHAO": ("、", False),
}


DEFAULT_ENGLISH_NER_MAPPING = {
    ",": "COMMA",
    ".": "PERIOD",
    "?": "QUESTIONMARK",
    "!": "EXLAMATIONMARK",
}

DEFAULT_CHINESE_NER_MAPPING = {
    "，": "C_COMMA",
    "。": "C_PERIOD",
    "？": "C_QUESTIONMARK",
    "！": "C_EXLAMATIONMARK",
    "、": "C_DUNHAO",
}


ALL_PUNCS = [
    c
    for c in range(sys.maxunicode)
    if unicodedata.category(chr(c)).startswith(("P", "Cc"))
]

currency_list = "|".join(
    [
        chr(c)
        for c in range(sys.maxunicode)
        if unicodedata.category(chr(c)).startswith(("Sc"))
    ]
)
CURRENCY = build_new_regex(
    "currency", r"(\{})\d+([.,]?\d*)*([A-Za-z]+)?".format(currency_list)
)

NUMBER = build_new_regex("number", r"[0-9]*[.]?[0-9]+[%]?")
# byte format
NUM_BYTE_FORMAT = "!H"
LENGTH_BYTE_FORMAT = "!I"

NUM_BYTE_LENGTH = 2
LENGTH_BYTE_LENGTH = 4



import logging
import re

from plane import CJK

logger = logging.getLogger(__name__)


def remove_brackets_text(input):
    """remove brackets and text inside

    Args:
        input (string): text to apply the regex func to

    """
    return re.sub(r"\([^()]*\)", " ", input)


def keep_only_latin_characters(input):
    """keep only latin characters

    Args:
        input (string): text to apply the regex func to

    """
    regex = re.compile("[^\u0020-\u024F]")
    return regex.sub("", input)


def chinese_split(input):
    """Split Chinese input by:
    - Adding space between every Chinese character. Note: English word will remain as original

    Args:
        input (string): text to apply the regex func to
    """

    regex = re.compile("(?P<%s>%s)" % (CJK.name, CJK.pattern), CJK.flag)
    result = ""
    start = 0
    try:
        for t in regex.finditer(input):
            result += input[start : t.start()].strip()
            result += (
                " "
                + " ".join(
                    [char for char in list(input[t.start() : t.end()]) if char != " "]
                )
                + " "
            )
            start = t.end()
        result += input[start:].strip()
    except TypeError as err:
        # mal row
        logger.warning(f"parsing data: {input} with error: {str(err)}")
    return result



import argparse
import os

from transformers import DistilBertForTokenClassification, DistilBertTokenizerFast


def upload_model(fine_tuned_model_path, fine_tuned_model_name, tokenizer_name):
    fine_tuned_model = DistilBertForTokenClassification.from_pretrained(
        fine_tuned_model_path
    )
    token = os.environ["HUGGINGFACE_TOKEN"]

    tokenizer = DistilBertTokenizerFast.from_pretrained(tokenizer_name)
    print(tokenizer)
    fine_tuned_model.push_to_hub(fine_tuned_model_name, use_auth_token=token)
    tokenizer.push_to_hub(fine_tuned_model_name, use_auth_token=token)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--fine_tuned_model_path", default="")
    parser.add_argument("--fine_tuned_model_name", default="distilbert_punctuator_en")
    parser.add_argument("--tokenizer_name", default="distilbert-base-uncased")
    args = parser.parse_args()

    upload_model(
        fine_tuned_model_path=args.fine_tuned_model_path,
        fine_tuned_model_name=args.fine_tuned_model_name,
        tokenizer_name=args.tokenizer_name,
    )



import logging
import re

from plane import replace
from plane.pattern import EMAIL, TELEPHONE
from tqdm import tqdm

from dbpunctuator.utils import (
    CURRENCY,
    CURRENCY_TOKEN,
    EMAIL_TOKEN,
    NUMBER,
    NUMBER_TOKEN,
    TELEPHONE_TOKEN,
    URL,
    URL_TOKEN,
)

tqdm.pandas()
logger = logging.getLogger(__name__)


def dataframe_data_cleaning(
    df, target_col, kept_punctuations, removed_punctuations, *special_cleaning_funcs
):
    """
    Clean up data in dataframe by removing all special characters except kept ones.
    """
    if special_cleaning_funcs:
        for func in special_cleaning_funcs:
            df[target_col] = df[target_col].progress_apply(lambda x: func(x))

    # replace email with <Special> and replace url with <URL>
    logger.info("replace email with <EMAIL>")
    df[target_col] = df[target_col].progress_apply(
        lambda x: replace(x, EMAIL, EMAIL_TOKEN)
    )

    logger.info("replace url with <URL>")
    df[target_col] = df[target_col].progress_apply(lambda x: replace(x, URL, URL_TOKEN))

    logger.info("replace currency with <CURRENCY>")
    df[target_col] = df[target_col].progress_apply(
        lambda x: replace(x, CURRENCY, CURRENCY_TOKEN)
    )

    logger.info("replace telephone with <TEL>")
    df[target_col] = df[target_col].progress_apply(
        lambda x: replace(x, TELEPHONE, TELEPHONE_TOKEN)
    )

    logger.info("replace number with <NUM>")
    df[target_col] = df[target_col].progress_apply(
        lambda x: replace(x, NUMBER, NUMBER_TOKEN)
    )

    translator = str.maketrans({key: None for key in removed_punctuations})
    space_translator = str.maketrans(
        {key: " {0} ".format(chr(key)) for key in kept_punctuations}
    )

    df[target_col] = df[target_col].progress_apply(
        lambda x: x.translate(translator).translate(space_translator).strip()
    )

    df.dropna(subset=[target_col])
    return df


def text_lines_cleaning(
    input_lines, kept_punctuations, removed_punctuations, *special_cleaning_funcs
):
    logger.info("clean up text file line by line.")
    logger.info("replace email with <EMAIL>")
    logger.info("replace url with <URL>")
    logger.info("replace currency with <CURRENCY>")
    logger.info("replace telephone with <TEL>")
    logger.info("replace number with <NUM>")

    pbar = tqdm(input_lines)
    for line in pbar:
        if special_cleaning_funcs:
            for func in special_cleaning_funcs:
                try:
                    line = func(line)
                except Exception as err:
                    logger.warning(f"error {str(err)} with func {func} for line {line}")
        line = replace(line, EMAIL, EMAIL_TOKEN)

        line = replace(line, URL, URL_TOKEN)

        line = replace(line, CURRENCY, CURRENCY_TOKEN)

        line = replace(line, TELEPHONE, TELEPHONE_TOKEN)

        line = replace(line, NUMBER, NUMBER_TOKEN)

        translator = str.maketrans({key: None for key in removed_punctuations})
        space_translator = str.maketrans(
            {key: " {0} ".format(chr(key)) for key in kept_punctuations}
        )

        yield line.translate(translator).translate(space_translator).strip()

    pbar.close()


def cleaning_validator(text, kept_punctuations, removed_punctuations):
    regex = re.compile(
        "[{}]".format("|".join(map(re.escape, [chr(p) for p in removed_punctuations])))
    )
    checking_result = regex.search(text)
    assert (
        checking_result is None
        or text[checking_result.span()[0] : checking_result.span()[1]]
        in kept_punctuations
    ), f"data cleaning for `{text}`` doesn't pass the validation with {checking_result}"
    return True



import logging

import torch
from pydantic import BaseModel
from transformers import DistilBertForTokenClassification, DistilBertTokenizerFast

from dbpunctuator.utils.utils import register_logger

logger = logging.getLogger(__name__)
register_logger(logger)

device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")


class TestingModelArguments(BaseModel):
    """Arguments pertaining to which model/config/tokenizer we are going to fine-tune from.

    Args:
        model_name(str): name or path of pre-trained model
        tokenizer_name(str): name of pretrained tokenizer
    """

    model_name: str
    tokenizer_name: str


class TestingModel:
    def __init__(self, arguments: TestingModelArguments) -> None:
        self.device = (
            torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")
        )
        self.tokenizer = DistilBertTokenizerFast.from_pretrained(
            arguments.tokenizer_name
        )
        self.classifer = DistilBertForTokenClassification.from_pretrained(
            arguments.model_name
        )

    def sample_output(self, inputs):
        tokenized_inputs = self.tokenizer(
            inputs,
            is_split_into_words=False,
            padding=True,
            truncation=True,
            return_offsets_mapping=True,
            return_tensors="pt",
        )
        logger.info(f"tokenized inputs: {tokenized_inputs}")
        self.tokenized_input_ids = tokenized_inputs["input_ids"].to(self.device)
        self.attention_mask = tokenized_inputs["attention_mask"].to(self.device)

        logits = self.classifer(self.tokenized_input_ids, self.attention_mask).logits
        if self.device.type == "cuda":
            argmax_preds = logits.argmax(dim=2).detach().cpu().numpy()
        else:
            argmax_preds = logits.argmax(dim=2).detach().numpy()
        logger.info(f"outputs of model {argmax_preds}")


if __name__ == "__main__":
    args = TestingModelArguments(
        model_name="distilbert-base-multilingual-cased",
        tokenizer_name="distilbert-base-multilingual-cased",
    )

    testing_model = TestingModel(args)
    test_texts = ["中文测试", "Chinese testing"]
    testing_model.sample_output(test_texts)
